//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#include <stdarg.h>
#ifdef _WIN32
	#include <windows.h>
#endif

#include <gtk/gtk.h>
#include <swfdec/swfdec.h>
#include <swfdec-gtk/swfdec-gtk.h>
#include "AnsiString.h"

#define CUSTOM_CONTROL_ID       1014

#define P_AUDIOENABLED          1000
#define P_PLAYING               1001
#define P_SPEED                 1002
#define P_FULLSCREENMODE        1003
#define P_INTERACTIVE           1004
#define P_ADVANCE               1005
#define P_ALIGNMENT             1006
#define P_PLAYERFOCUS           1007
#define P_MAXRUNTIME            1008
#define P_VARIABLES             1009
#define P_FRAMERATE             1010
#define P_SCALEMODE             1011
#define P_CACHESIZE             1012
#define P_SELECTION             1013
#define P_URL                   1014
#define P_BASEURL               1015

#define EVENT_ON_COMMAND        350
#define EVENT_ON_URL            351

INVOKER RemoteInvoker=0;
AnsiString buffered_text;

void on_command(SwfdecPlayer *player, gchar *command, gchar *parameter, gpointer user_data) {
    AnsiString data;
    data=(char *)command;
    if (parameter) {
        data+=(char *)"$";
        data+=parameter;
    }
    RemoteInvoker(user_data, INVOKE_FIREEVENT, (int)EVENT_ON_COMMAND, command);
}

void on_url(SwfdecPlayer *player,
            gchar        *url,
            gchar        *target,
            SwfdecBuffer *data,
            guint         header_count,
            GStrv        *header_names,
            GStrv        *header_values,
            gpointer      user_data)  {
    RemoteInvoker(user_data, INVOKE_FIREEVENT, (int)EVENT_ON_URL, url);
}

PREFIX int Invoke(void *context,int INVOKE_TYPE,...) {
    va_list arg;
    switch (INVOKE_TYPE) {
        case INVOKE_QUERYTEST:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id=va_arg(arg, int);
                va_end(arg);

                if (cls_id==CUSTOM_CONTROL_ID)
                    return E_ACCEPTED;
            }
            break;
        case INVOKE_SETEVENT:
            {
                if (!RemoteInvoker)
                    break;
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int event_id    =   va_arg(arg, int);
                char *str_buffer=   va_arg(arg, char *);
                int buf_len     =   va_arg(arg, int);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                SwfdecGtkWidget *swf=SWFDEC_GTK_WIDGET(context);
                SwfdecPlayer *player=0;
                if (swf)
                    player=swfdec_gtk_widget_get_player(swf);
                if (player) {
                    switch (event_id) {
                        case EVENT_ON_COMMAND:
                            g_signal_connect(player, "fscommand", G_CALLBACK (on_command),context);
                            return E_ACCEPTED;
                            break;
                        case EVENT_ON_URL:
                	        g_signal_connect (player, "launch", G_CALLBACK (on_url),context);
                            return E_ACCEPTED;
                            break;
                    }
                }
            }
            break;
        case INVOKE_SETPROPERTY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int prop_id     =   va_arg(arg, int);
                char *str_buffer=   va_arg(arg, char *);
                int buf_len     =   va_arg(arg, int);
                va_end(arg);
                AnsiString param(str_buffer);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                SwfdecGtkWidget *swf=SWFDEC_GTK_WIDGET(context);
                if (swf) {
                    SwfdecPlayer *player=swfdec_gtk_widget_get_player(swf);
                    switch (prop_id) {
                        case P_AUDIOENABLED:
                            swfdec_gtk_player_set_audio_enabled((SwfdecGtkPlayer *)player, param.ToInt());
                            break;
                        case P_PLAYING:
                            swfdec_gtk_player_set_playing((SwfdecGtkPlayer *)player, param.ToInt());
                            break;
                        case P_SPEED:
                            swfdec_gtk_player_set_speed((SwfdecGtkPlayer *)player, param.ToFloat());
                            break;
                        case P_FULLSCREENMODE:
                            swfdec_player_set_allow_fullscreen(player, param.ToInt());
                            break;
                        case P_INTERACTIVE:
                            // not use
                            break;
                        case P_ADVANCE:
                            if (player)
                                swfdec_player_advance(player, param.ToInt());
                            break;
                        case P_ALIGNMENT:
                            if (player)
                                swfdec_player_set_alignment(player, (SwfdecAlignment)param.ToInt());
                            break;
                        case P_PLAYERFOCUS:
                            if (player)
                                swfdec_player_set_focus(player, param.ToInt());
                            break;
                        case P_MAXRUNTIME:
                            if (player)
                                swfdec_player_set_maximum_runtime(player, param.ToInt());
                            break;
                        case P_VARIABLES:
                            if (player)
                                swfdec_player_set_variables(player, param.c_str());
                            break;
                        case P_FRAMERATE:
                            // readonly
                            break;
                        case P_SCALEMODE:
                            if (player)
                                swfdec_player_set_scale_mode(player, (SwfdecScaleMode)param.ToInt());
                            break;
                        case P_CACHESIZE:
                            if (player)
                                g_object_set(G_OBJECT(player), "cache-size", (gulong)param.ToInt(), NULL);
                            break;
                        case P_SELECTION:
                            // readonly
                            break;
                        case P_URL:
                            if (player)
                                swfdec_player_set_url(player, swfdec_url_new_from_input(param.c_str()));
                            break;
                        case P_BASEURL:
                            if (player)
                                swfdec_player_set_base_url(player, swfdec_url_new_from_input(param.c_str()));
                            break;

                    }
                }
                return E_NOERROR;
            }
            break;
        case INVOKE_GETPROPERTY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int prop_id     =   va_arg(arg, int);
                char **str_buffer=   va_arg(arg, char **);
                int *buf_len     =   va_arg(arg, int*);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                buffered_text=(char *)"";
                SwfdecGtkWidget *swf=SWFDEC_GTK_WIDGET(context);
                if (swf) {
                    SwfdecPlayer *player=swfdec_gtk_widget_get_player(swf);
                    switch (prop_id) {
                        case P_AUDIOENABLED:
                            if (player)
                                buffered_text=AnsiString((long)swfdec_gtk_player_get_audio_enabled((SwfdecGtkPlayer *)player));
                            break;
                        case P_PLAYING:
                            if (player)
                                buffered_text=AnsiString((long)swfdec_gtk_player_get_playing((SwfdecGtkPlayer *)player));
                            break;
                        case P_SPEED:
                            if (player)
                                buffered_text=AnsiString((double)swfdec_gtk_player_get_speed((SwfdecGtkPlayer *)player));
                            break;
                        case P_FULLSCREENMODE:
                            if (player)
                                buffered_text=AnsiString((long)swfdec_player_get_allow_fullscreen(player));
                            break;
                        case P_INTERACTIVE:
                            break;
                        case P_ADVANCE:
                            break;
                        case P_ALIGNMENT:
                            if (player)
                                buffered_text=AnsiString((long)swfdec_player_get_alignment(player));
                            break;
                        case P_PLAYERFOCUS:
                            if (player)
                                buffered_text=AnsiString((long)swfdec_player_get_focus(player));
                            break;
                        case P_MAXRUNTIME:
                            if (player)
                                buffered_text=AnsiString((long)swfdec_player_get_maximum_runtime(player));
                            break;
                        case P_VARIABLES:
                            if (player)
                                buffered_text=AnsiString((char *)swfdec_player_get_variables(player));
                            break;
                        case P_FRAMERATE:
                            if (player)
                                buffered_text=AnsiString((double)swfdec_player_get_rate(player));
                            break;
                        case P_SCALEMODE:
                            if (player)
                                buffered_text=AnsiString((long)swfdec_player_get_scale_mode(player));
                            break;
                        case P_CACHESIZE:
                            if (player) {
                                gulong res=0;
                                g_object_get(G_OBJECT(player), "cache-size", &res, NULL);
                                buffered_text=AnsiString((long)res);
                            }
                            break;
                        case P_SELECTION:
                            if (player)
                                buffered_text=(char *)swfdec_player_get_selection(player);
                            break;
                        case P_URL:
                            if (player) {
                                const SwfdecURL *url=swfdec_player_get_url(player);
                                if (url)
                                    buffered_text=(char *)swfdec_url_get_url(url);
                            }
                            break;
                        case P_BASEURL:
                            if (player) {
                                const SwfdecURL *url=swfdec_player_get_base_url(player);
                                if (url)
                                    buffered_text=(char *)swfdec_url_get_url(url);
                            } 
                            break;

                    }
                    *str_buffer=buffered_text.c_str();
                    *buf_len=buffered_text.Length();
                }
                return E_NOERROR;
            }
            break;
        case INVOKE_SETSTREAM:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                char *target    =   va_arg(arg, char *);
                int len_target  =   va_arg(arg, int);
                char *value     =   va_arg(arg, char *);
                int len_value   =   va_arg(arg, int);
                va_end(arg);
                return E_NOERROR;
            }
            break;
        case INVOKE_SETSECONDARYSTREAM:
            // not needed
            break;
        case INVOKE_CUSTOMMESSAGE:
            {
            }
            break;
        case INVOKE_CREATE:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                void **handler  =   va_arg(arg, void **);
                int *pack_type  =   va_arg(arg, int *);
                va_end(arg);
                *pack_type=(int)2;

                if (cls_id!=CUSTOM_CONTROL_ID)
                    break;

                GtkWidget *wrapped=swfdec_gtk_widget_new(swfdec_gtk_player_new (NULL)); 
                *handler=wrapped;

                return E_ACCEPTED;
            }
            break;
        case INVOKE_DESTROY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                void *handler  =   va_arg(arg, void *);
                va_end(arg);

                if (cls_id!=CUSTOM_CONTROL_ID)
                    break;

                return E_ACCEPTED;
            }
            break;
        case INVOKE_INITLIBRARY:
            va_start(arg, INVOKE_TYPE);
            RemoteInvoker     =   va_arg(arg, INVOKER);
            va_end(arg);

            swfdec_init();
            return E_NOERROR;

        case INVOKE_DONELIBRARY:
            return E_NOERROR;

        default:
            return E_NOTIMPLEMENTED;
    }
    return E_NOTIMPLEMENTED;
}


