//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#include <stdarg.h>
#include <gtk/gtk.h>

#include <stdlib.h>
#include <stdio.h>

#include <sys/types.h>
#include <sys/stat.h>

#ifdef _WIN32
	#include <windows.h>
	#include <process.h>
	#include <direct.h>
#else
	#include <sys/types.h>
	#include <sys/wait.h>
	#include <unistd.h>
#endif

extern "C" {
    #include "glade-window.h"
}

#include "AnsiString.h"


#define CUSTOM_CONTROL_ID       1005


#define P_OPERATION                 10003
#define P_PROJECTINDEX              10004
#define P_PALETTETYPE               10005
#define P_NEWPROJECT                10006
#define P_CLOSEPROJECT              10007


INVOKER RemoteInvoker=0;
AnsiString buffer;

PREFIX int Invoke(void *context,int INVOKE_TYPE,...) {
    va_list arg;
    switch (INVOKE_TYPE) {
        case INVOKE_QUERYTEST:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id=va_arg(arg, int);
                va_end(arg);

                if (cls_id==CUSTOM_CONTROL_ID)
                    return E_ACCEPTED;
            }
            break;
        case INVOKE_SETEVENT:
            {
                if (!RemoteInvoker)
                    break;
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int event_id    =   va_arg(arg, int);
                char *str_buffer=   va_arg(arg, char *);
                int buf_len     =   va_arg(arg, int);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

            }
            break;
        case INVOKE_SETPROPERTY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int prop_id     =   va_arg(arg, int);
                char *str_buffer=   va_arg(arg, char *);
                int buf_len     =   va_arg(arg, int);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                AnsiString s;
                if (buf_len)
                    s.LoadBuffer(str_buffer, buf_len);
	            GladeWindow *window=GLADE_WINDOW(context);

                switch (prop_id) {
                    case P_OPERATION:
                        if (s==(char *)"undo")
                            glade_undo();
                        else
                        if (s==(char *)"redo")
                            glade_redo();
                        else
                        if (s==(char *)"copy")
                            glade_copy();
                        else
                        if (s==(char *)"paste")
                            glade_paste();
                        else
                        if (s==(char *)"cut")
                            glade_cut();
                        else
                            return E_NOTIMPLEMENTED;
                        break;
                    case P_PROJECTINDEX:
                        {
                                void *project=get_project(s.ToInt());
                                if (project)
                                    select_project(window, project);
                        }
                        break;
                    case P_PALETTETYPE:
                        show_widgets(s.ToInt());
                        break;
                    case P_NEWPROJECT:
                        if (s.ToInt())
                            glade_window_new_project(window);
                        break;
                    case P_CLOSEPROJECT:
                        {
                            int proj_id=s.ToInt();
                            void *project=get_project(proj_id);
                            if (project)
                                close_project(window, project);
                        }
                        break;
                    default:
                        return E_NOTIMPLEMENTED;
                }
                
                return E_NOERROR;
            }
            break;
        case INVOKE_GETPROPERTY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int prop_id     =   va_arg(arg, int);
                char **str_buffer=   va_arg(arg, char **);
                int *buf_len     =   va_arg(arg, int*);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;


	            GladeWindow *window=GLADE_WINDOW(context);
                
                switch (prop_id) {
                    default:
                        return E_NOTIMPLEMENTED;
                }
                
                return E_NOERROR;
            }
            break;
        case INVOKE_SETSTREAM:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                char *target    =   va_arg(arg, char *);
                int len_target  =   va_arg(arg, int);
                char *value     =   va_arg(arg, char *);
                int len_value   =   va_arg(arg, int);
                va_end(arg);

                return E_NOERROR;
            }
            break;
        case INVOKE_SETSECONDARYSTREAM:
            // not needed
            break;
        case INVOKE_CUSTOMMESSAGE:
            {
                va_start(arg, INVOKE_TYPE);
                int     msg     =   va_arg(arg, int);
                int     cls_id  =   va_arg(arg, int);
                char   *target  =   va_arg(arg, char*);
                int     t_len   =   va_arg(arg, int);
                char   *value   =   va_arg(arg, char*);
                int     v_len   =   va_arg(arg, int);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                AnsiString s;
                AnsiString starget("gladetemp_");
                if (v_len)
                    s.LoadBuffer(value, v_len);

                if (t_len)
                    starget=starget+target;

                starget=starget+(char *)".glade";

	            GladeWindow *window=GLADE_WINDOW(context);

                switch (msg) {
                    case MSG_CUSTOM_MESSAGE1:
                        if (s.Length()) {
                            s.SaveFile(starget.c_str());
                            if (!glade_window_open_project(window,starget.c_str())) {
                                glade_window_new_project(window);
                            }
                            unlink(starget.c_str());
                        } else
                            glade_window_new_project(window);
                        break;
                    case MSG_CUSTOM_MESSAGE2:
                        {
                            int proj_id=s.ToInt();
                            void *project=get_project(proj_id);
                            s=(char *)"";
                            if (project) {
                                save_project(window, project, starget);
                                s.LoadFile(starget.c_str());
                                unlink(starget.c_str());
                            }
                            RemoteInvoker(context, INVOKE_SENDMESSAGE, (int)msg, (char*)"", (int)0,(char *)s.c_str(),(int)s.Length());
                        }
                        break;
                    case MSG_CUSTOM_MESSAGE3:
                        {
                            int proj_id=s.ToInt();
                            void *project=get_project(proj_id);
                            s=(char *)"-1";
                            if (project) {
                                s=(long)glade_is_saved(project);
                            }
                            RemoteInvoker(context, INVOKE_SENDMESSAGE, (int)msg, (char*)"", (int)0,(char *)s.c_str(),(int)s.Length());
                        }
                        break;
                    default:
                        return E_NOTIMPLEMENTED;
                }
                return E_NOERROR;
            }
            break;
        case INVOKE_CREATE:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                void **handler  =   va_arg(arg, void **);
                int *pack_type  =   va_arg(arg, int *);
                va_end(arg);
                *pack_type=(int)0;

                if (cls_id!=CUSTOM_CONTROL_ID)
                    break;

                GladeWindow *window=GLADE_WINDOW(glade_window_new());
                GtkWidget *widget=GTK_WIDGET(window);
                //glade_window_new_project(window);
                //show_widgets(2);
                *handler=widget;
				/*while (gtk_events_pending())
					gtk_main_iteration();*/
                return E_ACCEPTED;
            }
            break;
        case INVOKE_DESTROY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                void *handler  =   va_arg(arg, void *);
                va_end(arg);

                if (cls_id!=CUSTOM_CONTROL_ID)
                    break;
                
                //delete (Gtk::Widget*)handler;
                if (handler)
                    glade_window_config_save((GladeWindow *)handler);
                return E_ACCEPTED;
            }
            break;
        case INVOKE_INITLIBRARY:
            va_start(arg, INVOKE_TYPE);
            RemoteInvoker     =   va_arg(arg, INVOKER);
            va_end(arg);

            return E_NOERROR;

        case INVOKE_DONELIBRARY:
            return E_NOERROR;

        default:
            return E_NOTIMPLEMENTED;
    }
    return E_NOTIMPLEMENTED;
}


